from dataclasses import dataclass

import numpy as np

from pogema_toolbox.generators.generator_utils import maps_dict_to_yaml


@dataclass
class WarehouseConfig:
    wall_width: int = 5
    wall_height: int = 2
    walls_in_row: int = 5
    walls_rows: int = 5
    bottom_gap: int = 5
    horizontal_gap: int = 1
    vertical_gap: int = 3
    # wfi_instances: bool = True


def generate_warehouse(cfg: WarehouseConfig):
    height = cfg.vertical_gap * (cfg.walls_rows + 1) + cfg.wall_height * cfg.walls_rows
    width = cfg.bottom_gap * 2 + cfg.wall_width * cfg.walls_in_row + cfg.horizontal_gap * (cfg.walls_in_row - 1)

    grid = np.zeros((height, width), dtype=int)

    for row in range(cfg.walls_rows):
        row_start = cfg.vertical_gap * (row + 1) + cfg.wall_height * row
        for col in range(cfg.walls_in_row):
            col_start = cfg.bottom_gap + col * (cfg.wall_width + cfg.horizontal_gap)
            grid[row_start:row_start + cfg.wall_height, col_start:col_start + cfg.wall_width] = 1

    return '\n'.join(''.join('!' if cell == 0 else '#' for cell in row) for row in grid)


def generate_wfi_positions(grid_str, bottom_gap, vertical_gap):
    if vertical_gap == 1:
        raise ValueError("Cannot generate WFI instance with vertical_gap of 1.")

    grid = [list(row) for row in grid_str.strip().split('\n')]
    height = len(grid)
    width = len(grid[0])

    start_locations = []
    goal_locations = []

    for row in range(1, height - 1):
        if row % 3 == 0:
            continue
        for col in range(bottom_gap - 1):
            if grid[row][col] == '!':
                start_locations.append((row, col))
        for col in range(width - bottom_gap + 1, width):
            if grid[row][col] == '!':
                start_locations.append((row, col))

    if vertical_gap == 2:
        for row in range(1, height):
            for col in range(width):
                if grid[row][col] == '!' and grid[row - 1][col] == '#':
                    goal_locations.append((row, col))
    else:
        for row in range(height):
            for col in range(width):
                if grid[row][col] == '!':
                    if (row > 0 and grid[row - 1][col] == '#') or (row < height - 1 and grid[row + 1][col] == '#'):
                        goal_locations.append((row, col))

    return start_locations, goal_locations


def generate_wfi_warehouse(cfg: WarehouseConfig = WarehouseConfig()):
    grid = generate_warehouse(cfg)
    start_locations, goal_locations = generate_wfi_positions(grid, cfg.bottom_gap, cfg.vertical_gap)
    grid_list = [list(row) for row in grid.split('\n')]

    for s in start_locations:
        grid_list[s[0]][s[1]] = '@'
    for s in goal_locations:
        if grid_list[s[0]][s[1]] == '$':
            grid_list[s[0]][s[1]] = '!'
        else:
            grid_list[s[0]][s[1]] = '$'
    str_grid = '\n'.join([''.join(row) for row in grid_list])

    return str_grid


def main():
    maps_dict_to_yaml(f'warehouse.yaml', {'warehouse': generate_wfi_warehouse(WarehouseConfig(wall_width=10))})


if __name__ == '__main__':
    main()



import yaml


def maps_dict_to_yaml(filename, maps):
    with open(filename, 'w') as file:
        yaml.add_representer(str,
                             lambda dumper, data: dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|'))
        yaml.dump(maps, file)



import numpy as np
from numpy import random
from tabulate import tabulate

from pogema_toolbox.registry import ToolboxRegistry
from pogema_toolbox.views.view_utils import View, eval_logs_to_pandas, drop_na
from typing import Literal


class TabularView(View):
    print_results: bool = False
    type: Literal['tabular'] = 'tabular'
    table_format: str = 'psql'
    show_std: bool = True
    skip_zero_std: bool = True


class Aggregator:
    def __init__(self, estimator=np.mean, error_method="ci", error_level=1.96, round_digits=2):
        self.estimator = estimator
        self.error_method = error_method
        self.error_level = error_level
        self.round_digits = round_digits

    def __call__(self, vals):
        estimate = vals.agg(self.estimator) if not callable(self.estimator) else self.estimator(vals)
        interval_str = ""
        round_fmt = f"{{:.{self.round_digits}f}}"

        if self.error_method is None or len(vals) <= 1:
            interval_str = ""
        elif self.error_method == "sd":
            half_interval = vals.std() * self.error_level
            interval_str = f"± {round_fmt.format(half_interval)}"
        elif self.error_method == "ci":
            boots = self.bootstrap(vals, func=self.estimator)
            ci_lower, ci_upper = self._percentile_interval(boots, self.error_level)
            half_interval = (ci_upper - ci_lower) / 2
            interval_str = f"± {round_fmt.format(half_interval)}"
        else:
            raise KeyError(f"Unknown error_method {self.error_method}")

        return f"{round_fmt.format(estimate)} {interval_str}".strip()

    @staticmethod
    def bootstrap(data, n_boot=1000, func=np.mean):
        bs_samples = np.random.choice(data, replace=True, size=(n_boot, len(data)))
        bs_estimates = np.apply_along_axis(func, 1, bs_samples)
        return bs_estimates

    @staticmethod
    def _percentile_interval(data, percentile):
        lower_percentile = (100 - percentile) / 2
        upper_percentile = 100 - lower_percentile
        ci_lower, ci_upper = np.percentile(data, [lower_percentile, upper_percentile])
        return ci_lower, ci_upper


def preprocess_table(eval_configs, view):
    df = eval_logs_to_pandas(eval_configs)
    drop_na(df)

    metrics_keys = [key for key in eval_configs[0].get('metrics', []) if key in df.columns]
    drop_keys = list(view.drop_keys) if hasattr(view, 'drop_keys') else []
    group_by = [col for col in df.columns if col not in metrics_keys + drop_keys]

    # Define the error method and error level from the view configuration
    error_method = getattr(view, 'error_method', 'ci')  # Default to confidence interval
    error_level = getattr(view, 'error_level', 95)  # Default to 95% CI
    aggregator = Aggregator(error_method=error_method, error_level=error_level, round_digits=view.round_digits)

    # Define aggregation operations using the new Aggregator class
    agg_ops = {key: aggregator for key in metrics_keys}
    df_agg = df.groupby(by=group_by, as_index=False).agg(agg_ops).round(view.round_digits)

    # Drop specified keys, handle sorting, and renaming if required
    df_agg.drop(columns=drop_keys, errors='ignore', inplace=True)
    if hasattr(view, 'sort_by') and view.sort_by:
        df_agg.sort_values(by=view.sort_by, inplace=True)
    if hasattr(view, 'rename_fields') and view.rename_fields:
        df_agg.rename(columns=view.rename_fields, inplace=True)

    return df_agg


def process_table_view(results, view: TabularView):
    df = preprocess_table(results, view)

    table = tabulate(df, headers='keys', tablefmt=view.table_format)
    if view.print_results:
        ToolboxRegistry.info('\n' + table)


def generate_mock_data(num_results=25, ):
    results = []
    for i in range(num_results):
        result = {
            "metrics": {
                "avg_throughput": random.uniform(0.01, 0.25),  # Random float between 0.01 and 0.25
                "avg_num_agents_in_obs": random.uniform(1.4, 1.7),  # Random float between 1.4 and 1.7
                "runtime": random.uniform(29, 55)  # Random float between 29 and 55
            },
            "env_grid_search": {
                "map_name": f"mazes-seed-{i}-10x10",
                "num_agents": random.randint(10, 16),
                "seed": random.randint(0, 3)  # Random int between 0 and 1
            },
            "algorithm": "MATS-LP"
        }
        results.append(result)
    return results


def main():
    results = generate_mock_data()
    process_table_view(results, TabularView(round_digits=1, print_results=True, drop_keys=['seed', 'map_name']))
    process_table_view(results, TabularView(round_digits=2, print_results=True, drop_keys=[]))

    process_table_view(results, TabularView(round_digits=3, print_results=True, drop_keys=['seed', 'map_name', 'num_agents']))


if __name__ == '__main__':
    main()



from typing import Any, Dict, Generator, Tuple
import copy
from itertools import product


def generate_variants(spec: Dict) -> Generator[Tuple[Dict, Dict], None, None]:
    for variant, changes in _expand_grid_search(spec):
        if not _contains_unresolved_values(variant):
            yield changes, variant


def _expand_grid_search(spec: Dict) -> Generator[Tuple[Dict, Dict], None, None]:
    grid_params = _identify_grid_params(spec)
    if not grid_params:
        yield spec, {}
        return

    for combination in _generate_combinations(grid_params):
        variant, changes = _apply_combination(copy.deepcopy(spec), combination)
        yield variant, changes


def _identify_grid_params(spec: Dict, path=()) -> Dict:
    grid_params = {}
    for key, value in spec.items():
        current_path = path + (key,)
        if isinstance(value, dict):
            if "grid_search" in value:
                grid_params[current_path] = value["grid_search"]
            else:
                grid_params.update(_identify_grid_params(value, current_path))
    return grid_params


def _generate_combinations(grid_params: Dict) -> Generator[Dict, None, None]:
    keys, value_lists = zip(*grid_params.items())
    for values in product(*value_lists):
        yield dict(zip(keys, values))


def _apply_combination(spec: Dict, combination: Dict) -> Tuple[Dict, Dict]:
    changes = {}
    for path, value in combination.items():
        _set_dict_value(spec, path, value)
        changes[path] = value
    return spec, changes


def _set_dict_value(d: Dict, path: Tuple, value: Any):
    for key in path[:-1]:
        d = d.setdefault(key, {})
    d[path[-1]] = value


def _contains_unresolved_values(spec: Dict) -> bool:
    return any("grid_search" in value for _, value in _traverse_dict(spec) if isinstance(value, dict))


def _traverse_dict(d: Dict, path=()) -> Generator[Tuple[Tuple, Any], None, None]:
    for key, value in d.items():
        current_path = path + (key,)
        if isinstance(value, dict):
            yield from _traverse_dict(value, current_path)
        else:
            yield current_path, value


def main():
    environment_spec = {
        "name": "DecMAPF-v0",
        "grid_config": {
            "on_target": "restart",
            "max_episode_steps": 512,
            "map_name": {"grid_search": ["map-a", "map-b", "map-c", ]},
            "num_agents": {"grid_search": [8, 16, 32]},
            "seed": {"grid_search": [0, 1, 2]}
        }
    }

    # Generate and print the first 3 variant pairs to demonstrate the functionality
    variant_pairs = list(generate_variants(environment_spec))
    for changes, variant in variant_pairs[:3]:
        print(changes, variant)


if __name__ == '__main__':
    main()



from typing import Union, List

import pandas as pd
import yaml
from pydantic import BaseModel
from collections import defaultdict
from tabulate import tabulate

import json

from pogema_toolbox.registry import ToolboxRegistry


class View(BaseModel):
    drop_keys: list = ['seed']
    round_digits: int = 2
    rename_fields: dict = {"algorithm": "Algorithm",
                           "num_agents": "Number of Agents",
                           "avg_throughput": "Average Throughput",
                           "runtime": "Runtime (seconds)",
                           "max_episode_steps": "Episode Length"
                           }
    rename_algorithms: dict = {}
    sort_by: Union[str, List[str]] = None


def drop_na(df):
    # Drop columns with any NaN values
    cols_before_drop_na = set(df.columns)
    df = df.dropna(axis=1, how='any')
    cols_after_drop_na = set(df.columns)

    # Log dropped columns
    dropped_cols = cols_before_drop_na - cols_after_drop_na
    for col in dropped_cols:
        ToolboxRegistry.warning(f"Column '{col}' dropped due to missing values")

    return df


def eval_logs_to_pandas(eval_configs):
    data = {}
    for idx, config in enumerate(eval_configs):
        data[idx] = {**config['env_grid_search'], 'algorithm': config['algorithm']}

        # Adding metrics separately to skip possible lists of metrics (e.g. every step throughput)
        for key, value in config['metrics'].items():
            if isinstance(value, list):
                continue
            data[idx][key] = value
    return pd.DataFrame.from_dict(data, orient='index')


def load_from_folder(folder_path):
    eval_config_path = folder_path / (folder_path.name + '.yaml')
    # check if the file with name current_dir_name.name + '.yaml' exists
    assert eval_config_path.exists(), f'Config file {eval_config_path} does not exist'

    with open(eval_config_path, 'r') as f:
        evaluation_config = yaml.safe_load(f)

    results = []
    # Load results from *.json files
    for file in folder_path.glob('*.json'):
        with open(file, 'r') as f:
            results += json.load(f)
    return results, evaluation_config


def check_seeds(results):
    seed_data = defaultdict(lambda: defaultdict(list))
    algorithms = set()
    problem_count = 0  # Initialize problem counter

    # Populate seed_data and algorithms set
    for res in results:
        if 'env_grid_search' not in res:
            ToolboxRegistry.error("env_grid_search data missing")
            return

        env_data = res.get('env_grid_search', {})

        if 'map_name' not in env_data:
            ToolboxRegistry.error("No map_name in env_grid_search data")
            return

        if 'num_agents' not in env_data:
            ToolboxRegistry.error("No num_agents in env_grid_search data")
            return

        if 'seed' not in env_data:
            ToolboxRegistry.error("No seed in env_grid_search data")
            return

        map_name = res['env_grid_search']['map_name']
        num_agents = res['env_grid_search']['num_agents']
        seed = res['env_grid_search']['seed']
        algo = res['algorithm']
        algorithms.add(algo)
        seed_data[(map_name, num_agents, seed)][algo].append(res)

    # Prepare data for tabulate
    tabulate_data = []
    headers = ['Map Name', 'Num Agents', 'Seed'] + sorted(algorithms)

    for (map_name, num_agents, seed), algos in seed_data.items():
        row_issues = False
        row = [map_name, num_agents, seed]
        for algo in sorted(algorithms):
            if algo in algos:
                if len(algos[algo]) > 1:  # Duplicated seeds for this algo
                    row.append(f"x{len(algos[algo])}")
                    problem_count += 1  # Increment problem counter
                    row_issues = True
                else:
                    row.append("ok")  # Seed present and not duplicated
            else:
                row.append("missing")  # Seed missing for this algo
                problem_count += 1  # Increment problem counter
                row_issues = True

        if row_issues:  # Add row only if there are issues
            tabulate_data.append(row)

    # If problems detected, show error with tabulate
    if problem_count > 0:
        table = tabulate(tabulate_data, headers=headers, tablefmt='psql')
        ToolboxRegistry.error(f"Detected {problem_count} problems with seeds:\n" + table)
    else:
        ToolboxRegistry.success(f"Passed seeds consistency check")



import re
from dataclasses import dataclass

import svgwrite
import random
import math
import base64


@dataclass
class IsometricConfig:
    tile_width: int = 128
    tile_height: int = 32
    wall_height: int = 16
    view_angle: float = math.radians(70)
    stroke_color: str = "black"
    box_class_name: str = "cls-6"

    colors: tuple = (
        "#494C60",
        "#7C8398",
        "#99A2B3",
        "#737A5D",
        "#849E97",
        "#B0C2B0",
        "#B57168",
        "#D78A76",
        "#E8B39A",
        "#DBBDA1",
        "#DECAC1",
        "#E2DBDF",
    )


pattern = """
  ...#...#...#.........
  ...#.#.#...#####.#.#.
  ...#.#.....#.....#.#.
  ...###.....#.###.#.#.
  ...............#.#.#.
  ####.#####.###.#.#.#.
  .....#.......#.....#.
  .#.#######.#.###.#.#.
  ...#.....#.#.#...#.#.
  ####...#.#.#.#.#.#.#.
  ####...#.#.....#.#...
"""

pattern = [q.replace(" ", "") for q in pattern.split('\n') if q]
rows = len(pattern)
cols = len(pattern[0])

dwg = svgwrite.Drawing('01-warehouse.svg', profile='tiny')

min_x, max_x, min_y, max_y = float('inf'), float('-inf'), float('inf'), float('-inf')


def update_bounds(points):
    global min_x, max_x, min_y, max_y
    for x, y in points:
        min_x = min(min_x, x)
        max_x = max(max_x, x)
        min_y = min(min_y, y)
        max_y = max(max_y, y)


def create_iso_tile(x, y, color, cfg):
    cos_angle = math.cos(cfg.view_angle)
    sin_angle = math.sin(cfg.view_angle)
    points = [
        (x, y),
        (x + cfg.tile_width / 2 * cos_angle, y + cfg.tile_height / 2 * sin_angle),
        (x, y + cfg.tile_height * sin_angle),
        (x - cfg.tile_width / 2 * cos_angle, y + cfg.tile_height / 2 * sin_angle)
    ]
    update_bounds(points)
    return dwg.polygon(points, fill=color, stroke='black', stroke_width=1)


def create_iso_wall(x, y, cfg, wall_height=32, wall_color="#84A1AE", stroke_color='#333333'):
    cos_angle = math.cos(cfg.view_angle)
    sin_angle = math.sin(cfg.view_angle)
    half_width = cfg.tile_width / 2

    left = [
        (x, y + cfg.tile_height * sin_angle - wall_height * sin_angle),
        (x, y + cfg.tile_height * sin_angle),
        (x - half_width * cos_angle, y + cfg.tile_height / 2 * sin_angle),
        (x - half_width * cos_angle, y - wall_height * sin_angle + cfg.tile_height / 2 * sin_angle),
    ]
    update_bounds(left)
    left_wall = dwg.polygon(left, fill=wall_color, stroke=stroke_color, stroke_width=1)
    dwg.add(left_wall)

    top = [
        (x, y + cfg.tile_height * sin_angle - wall_height * sin_angle),
        (x + cfg.tile_width / 2 * cos_angle, y + cfg.tile_height / 2 * sin_angle - wall_height * sin_angle),
        (x, y - wall_height * sin_angle),
        (x - cfg.tile_width / 2 * cos_angle, y + cfg.tile_height / 2 * sin_angle - wall_height * sin_angle)
    ]
    update_bounds(top)
    top_wall = dwg.polygon(top, fill=wall_color, stroke=stroke_color, stroke_width=1)
    dwg.add(top_wall)

    front = [
        (x + half_width * cos_angle, y - wall_height * sin_angle + cfg.tile_height / 2 * sin_angle),
        (x, y - wall_height * sin_angle + cfg.tile_height * sin_angle),
        (x, y + cfg.tile_height * sin_angle),
        (x + half_width * cos_angle, y + cfg.tile_height / 2 * sin_angle),
    ]
    update_bounds(front)
    front_wall = dwg.polygon(front, fill=wall_color, stroke=stroke_color, stroke_width=1)
    dwg.add(front_wall)


def create_iso_flag(row, col, cfg, flag_height=25, flag_color="brown", ):
    cos_angle = math.cos(cfg.view_angle)
    sin_angle = math.sin(cfg.view_angle)

    base_x = (col - row) * cfg.tile_width / 2 * cos_angle + cols * cfg.tile_width / 2
    base_y = (col + row) * cfg.tile_height / 2 * sin_angle + cfg.tile_height / 2

    pole_top_x = base_x
    pole_top_y = base_y - flag_height * sin_angle

    flag_points = [
        (base_x, base_y),
        (pole_top_x, pole_top_y),
        (pole_top_x + 10, pole_top_y + 5),
        (pole_top_x, pole_top_y + 10)
    ]
    update_bounds(flag_points)
    flag_pole = dwg.line(start=(base_x, base_y), end=(pole_top_x, pole_top_y), stroke=cfg.stroke_color, stroke_width=2)
    flag = dwg.polygon(flag_points, fill=flag_color, stroke=cfg.stroke_color, stroke_width=1)

    small_tile_width = cfg.tile_width / 3
    small_tile_height = cfg.tile_height / 3
    offset_y = small_tile_height / 2
    small_tile_points = [
        (base_x, base_y - offset_y * sin_angle),
        (base_x + small_tile_width / 2 * cos_angle, base_y - offset_y * sin_angle + small_tile_height / 2 * sin_angle),
        (base_x, base_y - offset_y * sin_angle + small_tile_height * sin_angle),
        (base_x - small_tile_width / 2 * cos_angle, base_y - offset_y * sin_angle + small_tile_height / 2 * sin_angle)
    ]
    update_bounds(small_tile_points)
    small_tile = dwg.polygon(small_tile_points, fill=flag_color, stroke=cfg.stroke_color, stroke_width=1)

    return small_tile, flag_pole, flag


def read_robot_svg_as_data_url(file_path):
    with open(file_path, 'rb') as file:
        data = file.read()
    encoded_data = base64.b64encode(data).decode('utf-8')
    return f"data:image/svg+xml;base64,{encoded_data}"


def modify_robot_color(svg_content, cfg, new_color='#FFFFFF'):
    class_pattern = re.compile(r'(\.' + re.escape(cfg.box_class_name) + r'\s*\{\s*fill\s*:\s*)(#[0-9a-fA-F]{6})(\s*;?)\s*\}')
    modified_svg_content = class_pattern.sub(r'\1' + new_color + r'\3}', svg_content)
    return modified_svg_content


def insert_robot(dwg, robot_data_url, x, y, scale=0.75, offset=3):
    # Assuming the original size of the robot image is 50x50 pixels
    original_width = 50
    original_height = 50

    # Calculate the offsets to center the image
    offset_x = -original_width * scale / 2
    offset_y = -original_height * scale / 2

    # Create a group with the necessary transformations
    robot_group = dwg.g(transform=f"translate({x - offset}, {y + offset}) scale({scale})")
    robot_image = dwg.image(href=robot_data_url, insert=(offset_x, offset_y), size=(f"{original_width}px", f"{original_height}px"))
    robot_group.add(robot_image)
    dwg.add(robot_group)


def main():
    cfg = IsometricConfig()
    cos_angle = math.cos(cfg.view_angle)
    sin_angle = math.sin(cfg.view_angle)

    traversable_cells = [(row, col) for row in range(rows) for col in range(cols) if pattern[row][col] == '.']
    random.shuffle(traversable_cells)

    agent_positions = traversable_cells[:12]
    goal_positions = traversable_cells[16:16 + 12]
    agent_id = 0
    goal_id = 0
    robot_svg_path = 'Robot_v4.svg'

    with open(robot_svg_path, 'r') as robot_file:
        robot_svg_content = robot_file.read()

    for slice in range(rows + cols - 1):
        for row in range(rows):
            col = slice - row
            if col >= 0 and col < cols:
                iso_x = (col - row) * cfg.tile_width / 2 * cos_angle + cols * cfg.tile_width / 2
                iso_y = (col + row) * cfg.tile_height / 2 * sin_angle
                floor_tile = create_iso_tile(iso_x, iso_y, '#cccccc' if pattern[row][col] == '.' else '#666666', cfg)
                dwg.add(floor_tile)

    for slice in range(rows + cols - 1):
        for row in range(rows):
            col = slice - row
            if col >= 0 and col < cols:
                iso_x = (col - row) * cfg.tile_width / 2 * cos_angle + cols * cfg.tile_width / 2
                iso_y = (col + row) * cfg.tile_height / 2 * sin_angle
                if pattern[row][col] == '#':
                    create_iso_wall(iso_x, iso_y, cfg, wall_height=cfg.wall_height)
                if (row, col) in agent_positions:
                    modified_robot_svg_content = modify_robot_color(robot_svg_content, cfg, new_color=cfg.colors[agent_id])
                    robot_data_url = f"data:image/svg+xml;base64,{base64.b64encode(modified_robot_svg_content.encode()).decode()}"
                    insert_robot(dwg, robot_data_url, iso_x, iso_y)
                    agent_id += 1
                if (col, row) in goal_positions:
                    parts = create_iso_flag(col, row, cfg, flag_color=cfg.colors[-goal_id], )
                    goal_id += 1
                    for part in parts:
                        dwg.add(part)

    dwg['width'] = max_x - min_x
    dwg['height'] = max_y - min_y
    dwg['viewBox'] = f"{min_x} {min_y} {max_x - min_x} {max_y - min_y}"

    dwg.save()


if __name__ == '__main__':
    main()



from pathlib import Path
import yaml
import sys
from loguru import logger as base_logger

from pogema_toolbox.algorithm_config import AlgoBase


class ToolboxRegistry:
    _instances = {}
    _maps = None
    _algorithms = {}
    _envs = {}
    _logger = None
    _logger_config = {'level': 'INFO', 'sink': sys.stderr}

    # ----- Logging section -----
    @classmethod
    def setup_logger(cls, level="INFO", sink=None):
        sink = sink or sys.stderr
        format_string = "<level>{time:YYYY-MM-DD at HH:mm:ss} | Toolbox {level} | {message}</level>"
        cls._logger = base_logger.bind(name="ToolboxRegistry")
        cls._logger.remove()
        cls._logger.add(sink, level=level, format=format_string, colorize=True)
        cls._logger_config = {'level': level, 'sink': sink}

    @classmethod
    def _ensure_logger(cls):
        if cls._logger is None:
            cls.setup_logger()

    @classmethod
    def info(cls, message):
        cls._ensure_logger()
        cls._logger.info(message)

    @classmethod
    def debug(cls, message):
        cls._ensure_logger()
        cls._logger.debug(message)

    @classmethod
    def warning(cls, message):
        cls._ensure_logger()
        cls._logger.warning(message)

    @classmethod
    def error(cls, message):
        cls._ensure_logger()
        cls._logger.error(message)

    @classmethod
    def success(cls, message):
        cls._ensure_logger()
        cls._logger.success(message)

    # ----- State section -----
    @classmethod
    def get_state(cls):
        cls._ensure_logger()
        return {
            'maps': cls._maps,
            'algorithms': cls._algorithms,
            'envs': cls._envs,
            'logger_config': cls._logger_config
        }

    @classmethod
    def recreate_from_state(cls, state):
        cls._maps = state.get('maps', {})
        cls._algorithms = state.get('algorithms', {})
        cls._envs = state.get('envs', {})
        logger_config = state.get('logger_config', {})
        if logger_config:
            cls.setup_logger(level=logger_config.get('level', 'INFO'), sink=logger_config.get('sink', sys.stderr))
        cls.debug('Registry is recreated from state')

    # ----- Algorithms section -----
    @classmethod
    def register_algorithm(cls, name, make_func, config_make_func=None, preprocessing_make_func=None):
        if name in cls._algorithms:
            cls._logger.warning(f'Registering existing algorithm with name {name}')
        cls._algorithms[name] = make_func, config_make_func, preprocessing_make_func
        cls.debug(f'Registered algorithm with name {name}')

    @classmethod
    def create_algorithm(cls, algo_name, **kwargs):
        algo_make_func, config_make_func, _ = cls._algorithms[algo_name]
        cls.debug(f'Creating {algo_name} algorithm')
        if config_make_func:
            config = cls.create_algorithm_config(algo_name, **kwargs)
            return algo_make_func(config)
        return algo_make_func()

    @classmethod
    def create_algorithm_config(cls, algo_name, **kwargs):
        _, config_make_func, _ = cls._algorithms[algo_name]
        if config_make_func is None:
            cls.debug('Using default config - AlgoBase')
            return AlgoBase(**kwargs)
        cls.debug(f'Creating {algo_name} config')
        return config_make_func(**kwargs)

    @classmethod
    def create_algorithm_preprocessing(cls, env_to_wrap, algo_name, **kwargs):
        _, config_make_func, preprocessing_make_func = cls._algorithms[algo_name]
        if config_make_func is not None:
            algo_cfg = cls.create_algorithm_config(algo_name, **kwargs)
            cls.debug(f'Creating {algo_name} preprocessing')
            return preprocessing_make_func(env_to_wrap, algo_cfg)
        return preprocessing_make_func(env_to_wrap)

    # ----- Environments section -----
    @classmethod
    def register_env(cls, name, make_func, config_make_func=None):
        if name in cls._envs:
            cls._logger.warning(f'Registering existing environment with name {name}')
        cls._envs[name] = make_func, config_make_func
        cls.debug(f'Registered environment with name {name}')

    @classmethod
    def create_env(cls, env_name, **kwargs):
        env_make_func, config_make_func = cls._envs[env_name]

        if config_make_func:
            cls.debug(f'Creating {env_name} env using config')
            config = config_make_func(**kwargs)
            return env_make_func(config)
        cls.debug(f'Creating {env_name} env')
        return env_make_func()

    # ----- Maps section -----
    @classmethod
    def get_maps(cls):
        if cls._maps is None:
            cls._initialize_maps()
        return cls._maps

    @classmethod
    def register_maps(cls, maps):
        for map_name, map_grid in maps.items():
            if map_name in cls.get_maps():
                cls._logger.warning(f'Registering existing map with name {map_name}')
            cls._maps[map_name] = map_grid
            cls.debug(f'Registered map with name {map_name}')

    @classmethod
    def _initialize_maps(cls):
        maps_folder_path = Path(__file__).parent / "maps"
        cls._maps = {}

        for yaml_file_path in maps_folder_path.glob("*.yaml"):
            with open(yaml_file_path, "r") as f:
                try:
                    grids_content = yaml.safe_load(f)
                    if grids_content:  # Check if the YAML file is not empty
                        cls._maps.update(grids_content)
                except yaml.YAMLError as exc:
                    cls.error(f'Error loading YAML file {yaml_file_path}: {exc}')
        cls.debug(f'Registered {len(cls._maps)} maps')



import shutil
import tempfile

from pathlib import Path
import wandb


def initialize_wandb(evaluation_config, eval_dir, disable_wandb, project_name):
    mode = 'disabled' if disable_wandb else 'online'
    wandb.init(project=project_name, anonymous="allow", config=evaluation_config,
               mode=mode, job_type=eval_dir.stem, group='eval')


def save_evaluation_results(eval_dir):
    zip_path = f"{eval_dir}.zip"
    shutil.make_archive(str(eval_dir), 'zip', eval_dir)
    wandb.save(str(zip_path), )


def create_and_push_summary_archive(folder_names, base_path, project_name, archive_name='eval_summary'):
    """Create a zip archive and push it to wandb and return the download link."""

    with tempfile.TemporaryDirectory() as tempdir:
        for folder in folder_names:
            src = base_path / folder
            dst = Path(tempdir) / folder
            shutil.copytree(src, dst)

        archive_path = shutil.make_archive(str(base_path / archive_name), 'zip', tempdir)

        with wandb.init(project=project_name) as run:
            artifact = wandb.Artifact(
                name=archive_name,
                type='archive',
                description="Summary archive",
                metadata=dict(folder_names=folder_names)
            )
            artifact.add_file(archive_path)
            run.log_artifact(artifact)



import numpy as np
from dataclasses import dataclass

from pogema_toolbox.generators.generator_utils import maps_dict_to_yaml


@dataclass
class MazeRangeSettings:
    width_min: int = 5
    width_max: int = 9

    height_min: int = 5
    height_max: int = 9

    obstacle_density_min: float = 0.0
    obstacle_density_max: float = 1.0

    wall_components_min: int = 1
    wall_components_max: int = 8

    go_straight_min: float = 0.75
    go_straight_max: float = 0.85

    def sample(self, seed=None):
        rng = np.random.default_rng(seed)

        # Generate a sample for each attribute
        width = rng.integers(self.width_min, self.width_max + 1)
        height = rng.integers(self.height_min, self.height_max + 1)
        obstacle_density = rng.uniform(self.obstacle_density_min, self.obstacle_density_max)
        wall_components = rng.integers(self.wall_components_min, self.wall_components_max + 1)
        go_straight = rng.uniform(self.go_straight_min, self.go_straight_max)

        # Return a dictionary with the sampled values
        return {
            "width": width,
            "height": height,
            "obstacle_density": obstacle_density,
            "wall_components": wall_components,
            "go_straight": go_straight,
            "seed": seed,
        }


# Adapted from https://github.com/marmotlab/PRIMAL2/blob/main/Map_Generator.py
class MazeGenerator:

    @classmethod
    def array_to_string(cls, array_maze):
        result = []
        for line in array_maze:
            result.append("".join(['#' if x == 1 else '.' for x in line]))

        map_str = '\n'.join([''.join(row) for row in result])
        return map_str

    @classmethod
    def string_to_array(cls, string_maze):
        lines = string_maze.split('\n')
        array = np.array([[1 if char == '#' else 0 for char in line] for line in lines])
        return array

    @staticmethod
    def select_random_neighbor(x, y, maze_grid, maze_shape, rng, last_direction, go_straight):
        neighbor_coords = []
        probabilities = []
        if x > 1:
            neighbor_coords.append((y, x - 2))
            probabilities.append(
                go_straight if (y, x - 2) == (y + last_direction[0], x + last_direction[1]) else (1 - go_straight))
        if x < maze_shape[1] - 2:
            neighbor_coords.append((y, x + 2))
            probabilities.append(
                go_straight if (y, x + 2) == (y + last_direction[0], x + last_direction[1]) else (1 - go_straight))
        if y > 1:
            neighbor_coords.append((y - 2, x))
            probabilities.append(
                go_straight if (y - 2, x) == (y + last_direction[0], x + last_direction[1]) else (1 - go_straight))
        if y < maze_shape[0] - 2:
            neighbor_coords.append((y + 2, x))
            probabilities.append(
                go_straight if (y + 2, x) == (y + last_direction[0], x + last_direction[1]) else (1 - go_straight))

        if not neighbor_coords:
            return None, None, last_direction

        if all(prob == go_straight for prob in probabilities):  # Adjust probabilities if all are biased
            probabilities = [1 / len(probabilities)] * len(probabilities)
        else:
            total = sum(probabilities)
            probabilities = [prob / total for prob in probabilities]

        chosen_index = rng.choice(range(len(neighbor_coords)), p=probabilities)
        next_y, next_x = neighbor_coords[chosen_index]
        new_direction = (next_y - y, next_x - x)
        return next_x, next_y, new_direction

    @classmethod
    def generate_maze(cls, width, height, obstacle_density, wall_components, go_straight, seed=None):

        rng = np.random.default_rng(seed)
        assert width > 0 and height > 0, "Width and height must be positive integers"
        maze_shape = ((height // 2) * 2 + 3, (width // 2) * 2 + 3)
        density = int(
            maze_shape[0] * maze_shape[1] * obstacle_density // wall_components) if wall_components != 0 else 0

        maze_grid = np.zeros(maze_shape, dtype='int')
        maze_grid[0, :] = maze_grid[-1, :] = 1
        maze_grid[:, 0] = maze_grid[:, -1] = 1

        for i in range(density):
            x = rng.integers(0, maze_shape[1] // 2) * 2
            y = rng.integers(0, maze_shape[0] // 2) * 2
            maze_grid[y, x] = 1
            last_direction = (0, 0)  # Initial direction is null
            for j in range(wall_components):
                next_x, next_y, last_direction = MazeGenerator.select_random_neighbor(
                    x, y, maze_grid, maze_shape, rng, last_direction, go_straight
                )
                if next_x is not None and maze_grid[next_y, next_x] == 0:
                    maze_grid[next_y, next_x] = 1
                    maze_grid[next_y + (y - next_y) // 2, next_x + (x - next_x) // 2] = 1
                    x, y = next_x, next_y

        return cls.array_to_string(maze_grid[1:-1, 1:-1])

    @classmethod
    def draw(cls, mazes):
        if isinstance(mazes, str):
            mazes = [mazes]
        from matplotlib import pyplot as plt
        plt.ion()
        for maze in mazes:
            plt.imshow(cls.string_to_array(maze))
            plt.pause(0.1)
        plt.ioff()
        plt.show()

    @staticmethod
    def generate_maze_from_ranges(seed=None):
        pass


def generate_and_save_mazes(name_prefix, seed_range):
    test_mazes = {}
    # Determine the number of digits in the largest seed to set the zero-padding correctly
    max_digits = len(str(max(seed_range)))

    for seed in seed_range:
        settings = MazeRangeSettings().sample(seed)
        maze = MazeGenerator.generate_maze(**settings)
        # Apply dynamic zero-padding based on the number of digits in the largest seed
        map_name = f"{name_prefix}-seed-{str(seed).zfill(max_digits)}"
        test_mazes[map_name] = maze
    maps_dict_to_yaml(f'{name_prefix}.yaml', test_mazes)


def main():
    generate_and_save_mazes("validation-mazes", range(0, 128))
    generate_and_save_mazes("training-mazes", range(128, 128 + 2048))


if __name__ == "__main__":
    main()



from pathlib import Path

import yaml


def save_maps_to_single_yaml(filename, maps):
    with open(filename, 'w') as file:
        yaml.add_representer(str,
                             lambda dumper, data: dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|'))
        yaml.dump(maps, file)


def read_map_from_string(map_string):
    map_data = map_string.split('\n')
    return map_data


def split_map_into_tiles(map_data, tile_size):
    map_height = len(map_data)
    map_width = len(map_data[0])
    tiles = []

    for row in range(0, map_height, tile_size):
        for col in range(0, map_width, tile_size):
            tile = []
            for tile_row in range(tile_size):
                tile.append(map_data[row + tile_row][col:col + tile_size])
            tiles.append(tile)
    return ['\n'.join(tile) for tile in tiles]


def main():
    maps = {}
    for filename in Path("moving-ai-maps").glob("*.map"):

        with open(filename, 'r') as f:
            data = f.readlines()[4:]
            symbols = set()
            for idx, x in enumerate(data):
                data[idx] = x.strip()
                data[idx] = data[idx].replace("@", "#")

            for tile_idx, tile in enumerate(split_map_into_tiles(data, 64)):
                maps[filename.stem + '_' + str(tile_idx).zfill(2)] = tile

    save_maps_to_single_yaml("moving-ai-tiles.yaml", maps)


if __name__ == '__main__':
    main()



class ResultsHolder:
    """
    Holds and manages the results obtained during an episode.
    """

    def __init__(self):
        self.results = dict()

    def after_step(self, infos):
        if 'metrics' in infos[0]:
            self.results.update(**infos[0]['metrics'])

    def get_final(self):
        return self.results

    def __repr__(self):
        return str(self.get_final())



import numpy as np

from dataclasses import dataclass

from pogema_toolbox.generators.generator_utils import maps_dict_to_yaml


@dataclass
class MapRangeSettings:
    width_min: int = 17
    width_max: int = 21
    height_min: int = 17
    height_max: int = 21
    obstacle_density_min: float = 0.1
    obstacle_density_max: float = 0.3

    def sample(self, seed=None):
        rng = np.random.default_rng(seed)
        return {
            "width": rng.integers(self.width_min, self.width_max + 1),
            "height": rng.integers(self.height_min, self.height_max + 1),
            "obstacle_density": rng.uniform(self.obstacle_density_min, self.obstacle_density_max),
            "seed": seed
        }


def generate_map(settings):
    rng = np.random.default_rng(settings["seed"])
    width, height, obstacle_density = settings["width"], settings["height"], settings["obstacle_density"]
    map_data = [['.' for _ in range(width)] for _ in range(height)]
    total_tiles = width * height
    total_obstacles = int(total_tiles * obstacle_density)

    obstacles_placed = 0
    while obstacles_placed < total_obstacles:
        x = rng.integers(0, width)
        y = rng.integers(0, height)
        if map_data[y][x] == '.':
            map_data[y][x] = '#'
            obstacles_placed += 1

    return '\n'.join(''.join(row) for row in map_data)


def generate_and_save_maps(name_prefix, seed_range):
    test_maps = {}
    max_digits = len(str(max(seed_range)))
    settings_generator = MapRangeSettings()

    for seed in seed_range:
        settings = settings_generator.sample(seed)
        map_data = generate_map(settings)
        map_name = f"{name_prefix}-seed-{str(seed).zfill(max_digits)}"
        test_maps[map_name] = map_data

    maps_dict_to_yaml(f'{name_prefix}.yaml', test_maps)


def main():
    generate_and_save_maps("validation-random", range(0, 128))
    generate_and_save_maps("training-random", range(128, 128 + 512))


if __name__ == "__main__":
    main()
